html, body {
color: #DEE2EA;
background-color: #222733;
height: auto;
}
body {
font-family:"Lato";
padding: 50px;
}
td, th {
text-align: left;
vertical-align: top;
}
h1 {
font-family:"Lato";
font-size: 150px;
color: #929fba;
}
.cover {
padding-top: 20px;
}
.dq {
color: #0d7da5;
}
h2 {
font-size: 100px;
color: #0d7da5;
padding: 0; margin: 0;
margin-bottom: 140px;
}
h3 {
font-size: 25px;
color: #29b6f6;
padding: 0; margin: 0;
}
.product_code, .attr_name {
font-family:"Roboto Mono";
color: #DEE2EA;
}
h4.attr_name {
border-bottom: 1px solid #29b6f6;
}
.product_desc {
font-style: italic;
}
.product_tags {
font-weight: bold; 
color: #929fba;
padding-bottom: 10px;
}
.header {
width:100%;
padding: 0;
margin: 0; 
}
.header th {
padding: 0; margin: 0;
text-align: left;
color: #929fba;
font-size: 40px;
}
.header td {
padding: 0; margin: 0;
text-align: right;
}
.header_logo {
height: 40px;
}
.product_stats_diagram, .attr_stats_diagram {
width: 100%;
margin: 5px 0px;
}
.product_stats_diagram td, .attr_stats_diagram td {
font-family: "Lato";
text-align: right;
font-style: italic;
color: #000000;
padding: 5px;
}
.product_stats_legend {
font-family: "Lato";
font-style: italic;
color: #DEE2EA;
text-align: right;
padding-bottom: 10px;
}
.legend_item {
padding: 0 0.5em;
}
.product_stat {
padding-top: 20px;
padding-bottom: 30px;
page-break-inside: avoid;
}
.page_break {
page-break-after: always;
}
.page_break_spacer {
padding-top: 50px;
}
.data_stats {
width: 70%;
}
.data_stats_diagram {
width: 100%;
}
.data_stats_legend {
font-family: "Lato";
font-style: italic;
color: #DEE2EA;
}
.final_spacer {
display: block;
}
.details_table tbody th {
font-family:"Roboto Mono";
}
.details_table thead {
font-weight: bold;
color: #29b6f6;
}
.attr_field, .product_field {
width: 10%;
padding-top: 10px;
}
.attr_field_value {
width: 90%;
}
